# -*- coding: utf-8 -*-
import collections

import pandas as pd
from sqlalchemy import MetaData, Table, inspect
from sqlalchemy.sql import text

from dblinea.db_postgresql import DBPostgresql


class DBBase:

    _database = None
    _engine = None

    # TODO: OS dados de coneção com o banco devem vir de outro lugar!
    _available_databases = dict(
        {
            "gavo": {
                "ENGINE": "postgresql_psycopg2",
                "HOST": "desdb4.linea.gov.br",
                "PORT": "5432",
                "USER": "untrustedprod",
                "PASSWORD": "untrusted",
                "DATABASE": "prod_gavo",
            }
        }
    )

    def __init__(
        self,
        database="gavo",
        dbhost=None,
        dbname=None,
        dbuser=None,
        dbpass=None,
        dbport=None,
        dbengine="postgresql_psycopg2",
    ):
        # Se todas as variaveis de configuração forem None
        # Vai criar a conexão usando um dos _available_databases.
        # Default "gavo" ou o valor informado pelo usuario em database
        if all(x is None for x in [dbhost, dbname, dbuser, dbpass, dbport]):
            self.__set_database(database)
        else:
            # Se ao menos umas dessas variaveis
            # [dbhost, dbname, dbuser, dbpass, dbport]
            # For diferente de None, vai tentar criar uma conexão
            # usando os dados que o usuario passou.
            # Util para:
            # - Acessar outros bancos de dados que o usuario tenha acesso
            # - Conectar ao banco usando suas credenciais
            # - Para executar os Unit tests fora do ambiente.
            db_settings = {
                "ENGINE": dbengine,
                "HOST": dbhost,
                "PORT": "5432",
                "USER": dbuser,
                "PASSWORD": dbpass,
                "DATABASE": dbname,
            }

            self._database = DBPostgresql(db_settings)

    def __set_database(self, database):
        """Instancia a classe de Banco de dados.

        Este Metodo é utilizado quando é passado o
        parametro database na instancia da Classe.
        verifica se o database está na lista _available_databases

        Args:
            database (str): Nome do database como está no
            atributo _available_databases.

        Raises:
            ValueError: Database not available.
        """

        if database not in self._available_databases:
            raise ValueError("Database not available.")

        db_settings = self._available_databases[database]

        if db_settings["ENGINE"] == "postgresql_psycopg2":
            self._database = DBPostgresql(db_settings)

        # if db["ENGINE"] == "sqlite3":
        #     return DBSqlite(db)

        # if db_settings["ENGINE"] == "oracle":
        #     return DBOracle(db_settings)

    def available_databases(self):
        """Lista os bancos de dados disponiveis

        Returns:
            Lista de databases pre configurados na bibloteca.
            cada elemento da lista representa um db.

            exemplo:
            [{'config_name': 'gavo', 'dbname': '<database_name>',
            'host': '<database_host.linea.gov.br>', 'engine': 'postgresql_psycopg2'}]
        """

        dbs = list()
        for config_name in self._available_databases:
            dbs.append(
                dict(
                    {
                        "config_name": config_name,
                        "dbname": self._available_databases[config_name]["DATABASE"],
                        "host": self._available_databases[config_name]["HOST"],
                        "engine": self._available_databases[config_name]["ENGINE"],
                    }
                )
            )

        return dbs

    def get_engine(self):
        """Retorna uma sqlalchemy.Engine

        Cria ou retorna uma engine para o database solicitado na Instancia da DBBase.
        Mais informações sobre Engine em
        https://docs.sqlalchemy.org/en/14/core/connections.html#sqlalchemy.engine.Engine

        Returns:
            sqlalchemy.engine.Engine:
        """

        if self._engine is None:

            self._engine = self._database.get_engine()

        return self._engine

    def get_table(self, tablename, schema=None):
        """Retona uma instancia de sqlalchemy.schema.Table que representa uma tabela no database.
        https://docs.sqlalchemy.org/en/14/core/metadata.html#sqlalchemy.schema.Table

        Args:
            tablename (str): Nome da tabela sem o schema.
            schema (str, optional): Schema onde a tabela se encontra. Defaults to None.

        Returns:
            sqlalchemy.schema.Table: instancia de Table representando a tabela solicitada.
        """
        engine = self.get_engine()
        tbl = Table(tablename, MetaData(engine), autoload=True, schema=schema)
        return tbl

    def execute(self, stm):
        """Executa a query usando con.execute,
        recomendada para query de Delete, Update ou outras
        querys que não precisem de iteração com o resultado.

        Args:
            stm (statement): Query a ser executada, pode ser escrita em SqlAlchemy
            ou string no caso de string ela sera convertida para TextClause.

        Returns:
            CursorResult: [description]
        """
        with self.get_engine().connect() as con:
            return con.execute(stm)

    def fetchall(self, stm):
        """Executa a query e retorna todos os resultados em uma lista.

        Args:
            stm (statement): Query a ser executada, pode ser escrita em SqlAlchemy
            ou string no caso de string ela sera convertida para TextClause.

        Returns:
            list: Lista com os resultado no formato original do SqlAlchemy LegacyRow.
        """
        # Convert Raw sql to Sql Alchemy TextClause
        stm = self.raw_sql_to_stm(stm)

        with self.get_engine().connect() as con:
            queryset = con.execute(stm).fetchall()
            return queryset

    def fetchall_dict(self, stm):
        """Executa a query e retorna todos os resultados em uma lista de Dicts.
            exemplo:[{'col': 'value', ..., 'colN':'valueN'}]

        Args:
            stm (statement): Query a ser executada, pode ser escrita em SqlAlchemy
            ou string no caso de string ela sera convertida para TextClause.

        Returns:
            list: Resultado da query em uma lista de dicionários.
        """
        # Convert Raw sql to Sql Alchemy TextClause
        stm = self.raw_sql_to_stm(stm)

        with self.get_engine().connect() as con:
            queryset = con.execute(stm)

            rows = list()
            for row in queryset:
                rows.append(self.to_dict(row))

            return rows

    def fetchall_df(self, stm):
        """Executa a query usando Pandas e retorna um Dataframe com o resultado.

        Args:
            stm (statement): Query a ser executada, pode ser escrita em SqlAlchemy
            ou string no caso de string ela sera convertida para TextClause.

        Returns:
            Pandas.Dataframe: Dataframe com o resultado da query.
        """
        df = pd.read_sql(stm, con=self.get_engine())

        return df

    def fetchone(self, stm):
        """Executa a query retorna a primeira linha do resultado

        Args:
            stm (statement): Query a ser executada, pode ser escrita em SqlAlchemy
            ou string no caso de string ela sera convertida para TextClause.

        Returns:
            sqlalchemy.engine.row.LegacyRow: Primeira linha do resultado da query.
        """

        # Convert Raw sql to Sql Alchemy TextClause
        stm = self.raw_sql_to_stm(stm)

        with self.get_engine().connect() as con:
            queryset = con.execute(stm).fetchone()
            return queryset

    def fetchone_dict(self, stm):
        """Executa a query retorna a primeira linha do resultado convertida em Dict

        Args:
            stm (statement): Query a ser executada, pode ser escrita em SqlAlchemy
            ou string no caso de string ela sera convertida para TextClause.

        Returns:
            dict: Primeira linha do resultado da query.
        """
        # Convert Raw sql to Sql Alchemy TextClause
        stm = self.raw_sql_to_stm(stm)

        with self.get_engine().connect() as con:
            queryset = con.execute(stm).fetchone()

            if queryset is not None:
                return self.to_dict(queryset)
            else:
                return None

    def fetch_scalar(self, stm):
        """Retorna o valor da primeira coluna na primeira linha do resultado da query
        util para querys de count por exemplo, ou quando se quer apenas um unico valor.

        Args:
            stm (statement): Query a ser executada, pode ser escrita em SqlAlchemy
            ou string no caso de string ela sera convertida para TextClause.

        Returns:
            any: Valor da primeira coluna na primeira linha.
        """
        # Convert Raw sql to Sql Alchemy TextClause
        stm = self.raw_sql_to_stm(stm)

        with self.get_engine().connect() as con:
            return con.execute(stm).scalar()

    def to_dict(self, row):
        """Converte uma linha de resultado do SQLAlchemy queryset para Dict

        Args:
            row (sqlalchemy.engine.row.LegacyRow): Row retornada pelo execute.

        Returns:
            dict : Row convertida para Dict {colname: value, colname2: value2 ...}
        """
        return dict(collections.OrderedDict(row))

    def raw_sql_to_stm(self, stm):
        """Converte uma string raw sql para SqlAlchemy TextClause

        Args:
            stm (str): Query SQL em string. ex: Select * from tablename...

        Returns:
            TextClause: TextClause representando uma string SQL.
        """
        if isinstance(stm, str):
            return text(stm)
        return stm

    def get_table_columns(self, tablename, schema=None):
        """Retorna os nomes das colunas de uma tabela.

        Args:
            tablename (string): Nome da tabela sem schema.
            schema (string): Nome do schema ou None quando nao houver.

        Returns:
            columns (list): Colunas disponiveis na tabela
        """
        insp = inspect(self.get_engine())
        return [value["name"] for value in insp.get_columns(tablename, schema)]

    def describe_table(self, tablename, schema=None):
        """Retorna uma lista de dicionarios com nome e
        tipo das colunas de uma tabela.
        exemplo: [{"name": "", "type": ""}]

        Args:
            tablename (str): Nome da tabela sem schema.
            schema (str, optional): Nome do schema. Defaults to None.

        Returns:
            list: Lista de colunas com seu tipo
        """

        cols = list()

        insp = inspect(self.get_engine())
        for c in insp.get_columns(tablename, schema):
            cols.append(dict({"name": c["name"], "type": c["type"]}))

        return cols



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="dblinea",
    packages=find_packages(include=["dblinea"]),
    version="0.1.14",
    description="Python library to access LIneA's database from Python code. "
    "Useful to retrieve data inside LIneA's JupyterHub platform.",
    license="MIT",
    python_requires=">=3.8",
    setup_requires=["pytest-runner", "numpy"],
    install_requires=[
        "sqlalchemy>=1.4.25",
        "psycopg2-binary>=2.9.1",
        "numpy>=1.19.4",
        "pandas>=1.2.0",
        "requests>=2.23.0",
    ],
    tests_require=["pytest==4.4.1"],
    test_suite="tests",
    # Long description of your library
    long_description=long_description,
    long_description_content_type="text/markdown",
    # Either the link to your github or to your website
    url="https://github.com/linea-it/dblinea",
    # Link from which the project can be downloaded
    # download_url='https://github.com/linea-it/lineadb/archive/refs/tags/v0.1.0-alpha.tar.gz',
)



import requests
import json
from requests.auth import HTTPBasicAuth
import pandas as pd
from io import StringIO
from urllib.parse import urljoin


class ScienceServerApi:

    _token = None

    _enviroments = {
        "localhost": "http://localhost/dri/api/",
        "linea-dev": "https://scienceserver-dev.linea.gov.br/dri/api/",
        "linea": "https://scienceserver.linea.gov.br/dri/api/",
    }

    _base_api_url = None

    def __init__(self, token, host="linea"):

        self._base_api_url = self._enviroments[host]

        self._token = token

    def _generate_internal_name(self, name):

        return "".join(x if x.isalnum() else "_" for x in name)

    def _get_request(self, url, params):
        try:
            r = requests.get(
                url,
                params=params,
                headers=dict(
                    {
                        "Accept": "application/json",
                        "Content-Type": "application/json",
                        "Authorization": "Token {}".format(self._token),
                    }
                ),
            )

            if r.status_code == 200:
                return r.json()

            elif r.status_code == 403:
                # Não enviou as credenciais de usuario
                message = json.loads(str(r.text))["detail"]
                return dict(
                    {
                        "success": False,
                        "message": message,
                        "status_code": r.status_code,
                    }
                )

            elif r.status_code == 404:
                # Mensagem de erro pra Not Found.
                message = r.text
                return dict(
                    {
                        "success": False,
                        "message": message,
                        "status_code": r.status_code,
                    }
                )
            else:
                return dict(
                    {
                        "success": False,
                        "status_code": r.status_code,
                    }
                )

        except requests.exceptions.HTTPError as errh:
            message = "Http Error: {}".format(errh)
            return dict(
                {
                    "success": False,
                    "message": message,
                }
            )

        except requests.exceptions.ConnectionError as errc:
            message = "Connection Error: {}".format(errc)
            return dict(
                {
                    "success": False,
                    "message": message,
                }
            )

        except requests.exceptions.Timeout as errt:
            message = "Timeout Error: {}".format(errt)
            return dict(
                {
                    "success": False,
                    "message": message,
                }
            )

        except requests.exceptions.RequestException as err:
            message = "Request Error: {}".format(err)
            return dict(
                {
                    "success": False,
                    "message": message,
                }
            )

    def _post_request(self, url, payload):
        try:
            r = requests.post(
                url,
                data=json.dumps(payload),
                headers=dict(
                    {
                        "Accept": "application/json",
                        "Content-Type": "application/json",
                        "Authorization": "Token {}".format(self._token),
                    }
                ),
            )

            if r.status_code == 200:
                return r.json()

            elif r.status_code == 403:
                # Não enviou as credenciais de usuario
                message = json.loads(str(r.text))["detail"]
                return dict(
                    {
                        "success": False,
                        "message": message,
                        "status_code": r.status_code,
                    }
                )

            elif r.status_code == 404:
                # Mensagem de erro pra Not Found.
                message = r.text
                return dict(
                    {
                        "success": False,
                        "message": message,
                        "status_code": r.status_code,
                    }
                )
            else:
                return dict(
                    {
                        "success": False,
                        "status_code": r.status_code,
                    }
                )

        except requests.exceptions.HTTPError as errh:
            message = "Http Error: {}".format(errh)
            return dict(
                {
                    "success": False,
                    "message": message,
                }
            )

        except requests.exceptions.ConnectionError as errc:
            message = "Connection Error: {}".format(errc)
            return dict(
                {
                    "success": False,
                    "message": message,
                }
            )

        except requests.exceptions.Timeout as errt:
            message = "Timeout Error: {}".format(errt)
            return dict(
                {
                    "success": False,
                    "message": message,
                }
            )

        except requests.exceptions.RequestException as err:
            message = "Request Error: {}".format(err)
            return dict(
                {
                    "success": False,
                    "message": message,
                }
            )

    def _delete_request(self, url):
        try:
            r = requests.delete(
                url,
                headers=dict(
                    {
                        "Accept": "application/json",
                        "Authorization": "Token {}".format(self._token),
                    }
                ),
            )

            if r.status_code == 204:
                return True
            elif r.status_code == 400:
                return dict(
                    {
                        "success": False,
                        "message": "The server failed to perform the operation.",
                        "status_code": r.status_code,
                    }
                )
            elif r.status_code == 403:
                # Não enviou as credenciais de usuario
                message = json.loads(str(r.text))["detail"]
                return dict(
                    {
                        "success": False,
                        "message": message,
                        "status_code": r.status_code,
                    }
                )
            elif r.status_code == 404:
                # Mensagem de erro pra Not Found.
                message = json.loads(str(r.text))["detail"]
                return dict(
                    {
                        "success": False,
                        "message": message,
                        "status_code": r.status_code,
                    }
                )
            else:
                return dict(
                    {
                        "success": False,
                        "status_code": r.status_code,
                    }
                )

        except requests.exceptions.HTTPError as errh:
            message = "Http Error: {}".format(errh)
            return dict(
                {
                    "success": False,
                    "message": message,
                }
            )

        except requests.exceptions.ConnectionError as errc:
            message = "Connection Error: {}".format(errc)
            return dict(
                {
                    "success": False,
                    "message": message,
                }
            )

        except requests.exceptions.Timeout as errt:
            message = "Timeout Error: {}".format(errt)
            return dict(
                {
                    "success": False,
                    "message": message,
                }
            )

        except requests.exceptions.RequestException as err:
            message = "Request Error: {}".format(err)
            return dict(
                {
                    "success": False,
                    "message": message,
                }
            )

    def get_catalog(self, id):

        url = urljoin(self._base_api_url, "catalog/")
        params = dict({"id": int(id)})
        result = self._get_request(url, params)

        if "success" in result and result["success"] is False:
            return result

        if len(result) > 0:
            data = result[0]

            url = urljoin(self._base_api_url, "/target/#cv/%s" % str(data["id"]))

            catalog = dict(
                {
                    "id": data["id"],
                    "owner": data["owner"],
                    "date": data["prd_date"],
                    "internal_name": data["prd_name"],
                    "display_name": data["prd_display_name"],
                    "display_name": data["prd_display_name"],
                    "tbl_schema": data["tbl_schema"],
                    "tbl_name": data["tbl_name"],
                    "rows": data["tbl_rows"],
                    "url": url,
                }
            )

            return catalog
        else:
            # Retorna None se nenhum catalogo for encontrado
            # Pode aconter quando o id não existe ou se o usuario não tiver permissão ao id que está tentando acessar.
            return None

    def __register_target_list(
        self,
        name,
        data,
        cls="objects",
        releases=[],
        description=None,
        base64=False,
        mime="csv",
    ):
        url = urljoin(self._base_api_url, "import_target_list/")

        payload = dict(
            {
                "type": "catalog",
                "class": cls,
                "name": self._generate_internal_name(name),
                "displayName": name,
                "releases": releases,
                "isPublic": False,
                "description": description,
                "base64": base64,
                "mime": mime,
                "csvData": data,
            }
        )

        result = self._post_request(url, payload)

        if "success" in result and result["success"] is False:
            return result

        if result["success"] is True:
            # Importou com sucesso
            catalog = self.get_catalog(result["product"])
            return catalog

    def target_list_from_list(
        self,
        name,
        data,
        cls="objects",
        releases=[],
        description=None,
    ):

        # Criar um dataframe e converter para string csv.
        df = pd.DataFrame(data)

        f = StringIO()
        df.to_csv(
            f,
            sep=";",
            header=True,
            index=False,
        )
        str_csv = f.getvalue()

        return self.__register_target_list(
            name, str_csv, cls, releases, description, base64=False, mime="csv"
        )

    def target_list_from_df(
        self,
        name,
        df,
        cls="objects",
        releases=[],
        description=None,
    ):

        # Converter para string csv.
        f = StringIO()
        df.to_csv(
            f,
            sep=";",
            header=True,
            index=False,
        )
        str_csv = f.getvalue()

        return self.__register_target_list(
            name, str_csv, cls, releases, description, base64=False, mime="csv"
        )

    def remove_target_list(self, id):

        url = urljoin(self._base_api_url, "catalog/%s/" % int(id))

        result = self._delete_request(url)

        if result is True:
            return dict(
                {"success": True, "message": "Target List successfully removed"}
            )
        else:
            return result


# from random import randint

# token = "c7aef9b2fbe8e9b5dea02e456c8075877b1dc841"
# name = "Teste %s" % randint(10, 50)
# data = "RA;DEC;Mag_g\n10;20;24.5\n25;15;22.7"
# data = [{"RA": 10, "DEC": 20, "Mag_g": 24.5}, {"RA": 25, "DEC": 15, "Mag_g": 22.7}]

# ss = ScienceServerApi(token, host="localhost")

# catalog = ss.target_list_from_list(name=name, data=data)
# print(catalog)

# df = pd.DataFrame(data)
# catalog = ss.target_list_from_df(name=name, df=df)
# print(catalog)

# catalog = ss.get_catalog(26)
# print(catalog)

# catalog = ss.remove_target_list(100)
# print(catalog)



from sqlalchemy import create_engine
from sqlalchemy.dialects import postgresql
from sqlalchemy.pool import NullPool
from sqlalchemy.sql import and_, or_, text


class DBPostgresql:

    __engine_name = "postgresql_psycopg2"

    __dialect = postgresql

    # Bulk Insert (se é possivel fazer inserts em blocos.)
    __bulk_insert = True

    def __init__(self, db_settings):
        self.db_settings = db_settings

    def get_db_uri(self):

        database = self.db_settings.get("DATABASE", None)
        if database is not None:
            uri = (
                "postgresql+psycopg2://%(username)s:%(password)s@%(host)s:%(port)s/%(database)s"
            ) % {
                "username": self.db_settings.get("USER"),
                "password": self.db_settings.get("PASSWORD"),
                "host": self.db_settings.get("HOST", "localhost"),
                "port": self.db_settings.get("PORT", "5432"),
                "database": database,
            }
        else:
            uri = (
                "postgresql+psycopg2://%(username)s:%(password)s@%(host)s:%(port)s"
            ) % {
                "username": self.db_settings.get("USER"),
                "password": self.db_settings.get("PASSWORD"),
                "host": self.db_settings.get("HOST", "localhost"),
                "port": self.db_settings.get("PORT", "5432"),
            }
        return uri

    def get_engine(self):
        return create_engine(self.get_db_uri(), poolclass=NullPool)

    def get_engine_name(self):
        return self.__engine_name

    def get_dialect(self):
        return self.__dialect

    def accept_bulk_insert(self):
        return self.__bulk_insert

    def get_condition_square(
        self, lowerleft, upperright, property_ra="ra", property_dec="dec"
    ):

        raul = float(lowerleft[0])
        decul = float(upperright[1])
        ul = "{%s, %s}" % (raul, decul)

        raur = float(upperright[0])
        decur = float(upperright[1])
        ur = "{%s, %s}" % (raur, decur)

        ralr = float(upperright[0])
        declr = float(lowerleft[1])
        lr = "{%s, %s}" % (ralr, declr)

        rall = float(lowerleft[0])
        decll = float(lowerleft[1])
        ll = "{%s, %s}" % (rall, decll)

        # ul, ur, lr, ll
        stm = "q3c_poly_query(%s, %s, '{ %s, %s, %s, %s}')" % (
            property_ra,
            property_dec,
            ul,
            ur,
            lr,
            ll,
        )

        return and_(text(stm)).self_group()

    # def get_raw_sql_limit(self, offset, limit):
    #     return "OFFSET %s LIMIT %s" % (offset, limit)

    # def get_create_auto_increment_column(self, table, column_name, schema=None):
    #     raise Exception(
    #         "Method not implemented 'get_create_auto_increment_column'")

    # def get_create_auto_increment_column(self, table, column_name, schema=None):
    #     table_name = table
    #     if schema is not None and schema is not "":
    #         table_name = "%s.%s" % (schema, table)

    #     sql = list()
    #     sql.append("CREATE INDEX %(table)s_%(column)s_idx ON %(table)s USING btree (%(column)s);" % {
    #         "table": table_name,
    #         "column": column_name})

    #     return sql

    # def get_raw_sql_column_properties(self, table, schema=None):
    #     sql = "SELECT column_name, data_type FROM information_schema.columns WHERE table_name = '%s'" % table
    #     if schema:
    #         sql += " AND table_schema = '%s'" % schema
    #     return sql

    # def get_raw_sql_table_rows(self, table, schema=None):
    #     if schema:
    #         sql = "SELECT reltuples FROM pg_class WHERE oid = '%s.%s'::regclass" % (
    #             schema, table)
    #     else:
    #         sql = "SELECT reltuples FROM pg_class WHERE relname='%s'" % table
    #     return sql

    # def get_raw_sql_size_table_bytes(self, table, schema=None):
    #     sql = "SELECT pg_total_relation_size(relid) as size_in_bytes  FROM pg_catalog.pg_statio_user_tables WHERE relname = '%s'" % table
    #     if schema:
    #         sql += " AND schemaname='%s'" % schema
    #     return sql

    # def get_raw_sql_number_columns(self, table, schema=None):
    #     where = "WHERE table_name = '%s'" % table
    #     if schema:
    #         where += " AND table_schema = '%s'" % schema
    #     sql = "SELECT count(*) as column_count FROM information_schema.columns %s GROUP by table_name order by column_count desc" % where
    #     return sql
